"""Studying energy startups

>>> wc = word_cloud(['here', 'are', 'some', 'words'])
>>> type(wc.to_image())
<class 'PIL.Image.Image'>


"""

from functools import partial
import pandas as pd
from dol import wrap_kvs


from guise.util import (
    url_to_html,
    get_html_store,
    word_cloud,
)

from guise.word_scoring import (
    mk_word_score_base_color_func,
    sentiment_score,
    term_filtered_html2text,
)

EX_DFLT_HTML_STORE_DIR = '~/odat/energy_startups/html'
url_to_html = partial(url_to_html, EX_DFLT_HTML_STORE_DIR)
get_html_store = get_html_store(get_html_store, EX_DFLT_HTML_STORE_DIR)


def get_data():
    from guise.util import proj_files

    data_files = proj_files / 'examples' / 'energy_startups' / 'data'
    df = pd.read_excel(data_files / 'energy5.xlsx')
    df.columns = list(map(str.lower, df.columns))
    return df


def name_url_iter(df):
    for _, row in df.iterrows():
        technology, company, country = map(
            lambda x: row[x].strip(), ('technology', 'company', 'country')
        )
        company = row.company.replace(' ', '_').strip()
        yield f'{technology},{company},{country}', row['website']


def scrape_and_store(
    name_and_url_iterable, html_store=None, overwrite=False, verbose=True
):
    html_store = get_html_store(html_store)
    for name, url in name_and_url_iterable:
        if overwrite and name in html_store:
            if verbose:
                print(f'Skipping because I already have: {name=}: {url=}')
        else:
            if verbose:
                print(f'Processing {url=}')
            html = url_to_html(url)
            if html is not None:
                html_store[f'{name}.html'] = html


def get_text_store(html_store=None, html2text=term_filtered_html2text):
    _html_2_text_store = wrap_kvs(
        key_of_id=lambda x: x[: -len('.html')],
        id_of_key=lambda x: x + '.html',
        obj_of_data=html2text,
    )
    return _html_2_text_store(get_html_store(html_store))



"""Semantic fingerprinting

>>> from guise import word_cloud
>>> wc = word_cloud(['here', 'are', 'some', 'words'])
>>> type(wc.to_image())
<class 'PIL.Image.Image'>
"""

from guise.util import (
    proj_files,
    google_search_html,
    google_results_urls,
    url_to_html,
    all_links_of_html,
)
from guise.nlp import (
    html_tokens,
    term_filtered_html2text,
    stem_based_word_mapping,
    stem_based_word_count,
    DFLT_INCLUDE_TERMS,
    DFLT_EXCLUDE_TERMS,
)
from guise.word_scoring import (
    sentiment_score,
    mk_word_score_base_color_func,
)
from guise.tools import google_results_toks, word_cloud



"""Tools to get scores from words"""

from functools import partial

# ---------------------------------------------------------------------------------------
# sentiment scoring

from functools import lru_cache

DLFT_TEXT_CLASSIFIER_LANG = 'en-sentiment'


@lru_cache
def _get_text_classifier(lang=DLFT_TEXT_CLASSIFIER_LANG):
    from flair.models import TextClassifier

    return TextClassifier.load(lang)


def _sentiment_score_object(string, lang=DLFT_TEXT_CLASSIFIER_LANG):
    from flair.data import Sentence

    sia = _get_text_classifier(lang)

    sentence = Sentence(string)
    sia.predict(sentence)
    return sentence.labels[0]


def sentiment_score(string):
    score = _sentiment_score_object(string)
    if score.value == 'NEGATIVE':
        return -score.score
    elif score.value == 'POSITIVE':
        return score.score
    else:
        raise ValueError(f"Didn't know score.value could be {score.value}")


# ---------------------------------------------------------------------------------------
# score to color


def _score_to_index(score, max_index, min_score=-1, max_score=1):
    return int(max_index * (score - min_score) / (max_score - min_score))


def _get_red_to_blue_colors(n=202):
    from colour import Color

    red = Color('red')
    colors = list(map(str, red.range_to(Color('blue'), n)))[1:-1]
    return colors


def mk_word_score_base_color_func(word_to_score=sentiment_score, colors=None):
    colors = colors or _get_red_to_blue_colors()

    score_to_index = partial(
        _score_to_index, max_index=len(colors), min_score=-1, max_score=1
    )

    def color_func(word, font_size, position, orientation, random_state=None, **kwargs):
        score = word_to_score(word)
        return colors[score_to_index(score)]

    return color_func



"""Natural language processing tools"""

from typing import Mapping, Any, Iterable, Union, Callable
from collections import Counter, defaultdict

from html2text import html2text
from idiom import word_frequencies
from lexis import Lemmas

from guise.util import STOPWORDS

Word = str
Words = Iterable[Word]
WordCount = Mapping[Word, int]
WordStats = Mapping[Word, Any]
WordMap = Mapping[Word, Word]
StemmerSpec = Union[Callable, str, None]
# ---------------------------------------------------------------------------------------
# tfidf scoring


_cache = {}


DFLT_INCLUDE_TERMS = 'most_frequent_english'
DFLT_EXCLUDE_TERMS = STOPWORDS


def get_include_terms(include=DFLT_INCLUDE_TERMS) -> Words:
    """allowed_terms = 'most_frequent_english' or 'wordnet_lemmas'"""
    include = include or 'most_frequent_english'
    if include == 'most_frequent_english':
        if 'most_frequent_english' not in _cache:
            _cache['most_frequent_english'] = set(dict(word_frequencies()))
        include = _cache['most_frequent_english']
    elif include == 'wordnet_lemmas':
        if 'wordnet_lemmas' not in _cache:
            _cache['wordnet_lemmas'] = set(Lemmas())
        include = _cache['wordnet_lemmas']
    return include


def html_tokens(html: str, include=DFLT_INCLUDE_TERMS, exclude=STOPWORDS) -> Words:
    toks = map(str.strip, html2text(html).split(' '))
    if include is not None:
        include = get_include_terms(include)
        include = set(include) - (exclude or {})
        return filter(include.__contains__, toks)
    elif exclude is not None:
        return filter(lambda x: x not in exclude, toks)
    else:
        return toks


def term_filtered_html2text(
    html, include=DFLT_INCLUDE_TERMS, exclude=DFLT_EXCLUDE_TERMS, sep=' ; '
) -> str:
    """allowed_terms = 'most_frequent_english' or 'wordnet_lemmas'"""
    return sep.join(html_tokens(html, include, exclude))


def stem_based_word_mapping(
    words: Words, stemmer: StemmerSpec = 'lancaster'
) -> WordMap:
    """Returns a dict of {word: replace_with_word, ...} mapping based on stemming.

    stemming is great to reduce the number of words by replacing several


    The words (or word counts) are stemmed, then for each stem, the word with the
    highest count gets to represent the others.

    >>> _words = ['happy', 'happier', 'happiest', 'happily', 'sad', 'sadly']
    >>> word_count = Counter(_words)
    >>> stem_based_word_mapping(word_count)  # doctest: +NORMALIZE_WHITESPACE
    {'happy': 'happy',
     'happier': 'happy',
     'happiest': 'happiest',
     'happily': 'happy',
     'sad': 'sad',
     'sadly': 'sad'}
    """
    if not callable(stemmer):
        from nltk.stem import PorterStemmer
        from nltk.stem import LancasterStemmer

        if stemmer is None:
            stemmer = lambda word: word
        elif isinstance(stemmer, str):
            if stemmer.lower().startswith('porter'):
                stemmer = PorterStemmer().stem
            elif stemmer.lower().startswith('lancaster'):
                stemmer = LancasterStemmer().stem
            else:
                raise ValueError(f'Unknown stemmer value: {stemmer}')
        else:
            raise TypeError(f'Unknown stemmer type: {stemmer}')

    if isinstance(words, Mapping):
        word_count = words
    else:
        word_count = Counter(words)

    word_of_stem = defaultdict(list)
    for word, count in word_count.items():
        word_of_stem[stemmer(word)].append((count, word))

    def sort_key(count_and_word):
        count, word = count_and_word
        # favor the highest count, if not, the shortest word, if not lexicographic
        return -count, len(word), word

    word_of_stem = {
        stem: sorted(count_and_words, key=sort_key)[0][1]
        for stem, count_and_words in word_of_stem.items()
    }

    return {word: word_of_stem[stemmer(word)] for word in word_count}


def map_words_of_word_count(word_count: WordCount, word_map: WordMap) -> WordCount:
    """Recompute a word count based on a (complete) mapping of it's words (keys)

    >>> assert (dict(
    ... map_words_of_word_count(
    ...     {'one': 1, 'ones': 2, '1': 3, 'two': 4},
    ...     {'one': 'one', 'ones': 'one', '1': 'one', 'two': 'two'}))
    ... == {'one': 6, 'two': 4}
    ... )

    """
    c = Counter()
    for word, count in word_count.items():
        c.update({word_map[word]: count})
    return c


def stem_based_word_count(
    words: Words, stemmer: StemmerSpec = 'lancaster'
) -> WordCount:
    """From an iterable of words, get a WordCount mapping which uses a stem-based map

    >>> words = ['happy', 'happier', 'happiest', 'happily', 'sad', 'sadly']
    >>> assert (dict(stem_based_word_count(words))
    ... == {'happy': 3, 'happiest': 1, 'sad': 2})

    """
    word_count = Counter(words)
    word_map = stem_based_word_mapping(word_count, stemmer)
    return map_words_of_word_count(word_count, word_map)



"""Tools"""
from typing import Mapping
from functools import partial
from itertools import chain
from collections import Counter
from inspect import signature

from wordcloud import WordCloud

from guise.util import google_search_html, google_results_urls, url_to_html
from guise.nlp import html_tokens, DFLT_INCLUDE_TERMS, DFLT_EXCLUDE_TERMS

DFLT_URL_TO_HTML_KWARGS = (('timeout', 20),)

# TODO: Can be accelerated significantly by async
def google_results_toks(
    q,
    *,
    num=30,
    lr='lang_en',
    include=DFLT_INCLUDE_TERMS,
    exclude=DFLT_EXCLUDE_TERMS,
    verbose=True,
    url_to_html_kwargs: dict = DFLT_URL_TO_HTML_KWARGS,
):
    """A generator of tokens """
    # get the google results html for query q
    results_html = google_search_html(q, num=num, lr=lr)

    # parse out the results urls
    results_urls = google_results_urls(results_html)

    # get the landing page (html) of each one of those result urls
    htmls = [url_to_html(url, **dict(url_to_html_kwargs)) for url in results_urls]

    # make a list of urls whose htmls couldn't be acquired
    # (status_code>200, or other kind of problem)
    problematic_urls = [results_urls[i] for i, html in enumerate(htmls) if html is None]

    # filter out the "bad" results
    htmls = list(filter(None, htmls))
    if problematic_urls and verbose:
        print('There were some problematic urls:')
        print(*problematic_urls, sep='\n')

    # return an iterator of tokens (words/terms) extracted from these htmls
    tokenizer = partial(html_tokens, include=include, exclude=exclude)
    return chain.from_iterable(map(tokenizer, htmls))


def word_cloud(
    words,
    save_filepath=None,
    *,
    color_func=None,
    stopwords=None,
    width=600,
    height=600,
    random_state=42,  # set to None to get a different image every time
    font_path=None,
    margin=2,
    ranks_only=None,
    prefer_horizontal=0.9,
    mask=None,
    scale=1,
    max_words=200,
    min_font_size=4,
    background_color='black',
    max_font_size=None,
    font_step=1,
    mode='RGB',
    relative_scaling='auto',
    regexp=None,
    collocations=True,
    colormap=None,
    normalize_plurals=True,
    contour_width=0,
    contour_color='black',
    repeat=False,
    include_numbers=False,
    min_word_length=0,
    collocation_threshold=30,
):
    word_cloud_kwargs = {
        k: v for k, v in locals().items() if k in set(signature(WordCloud).parameters)
    }
    if isinstance(words, str):
        # simple tokenization of string into words
        words = list(map(str.strip, words.split(' ')))
    if isinstance(words, Mapping):
        # We already have a {word: frequency, ...} mapping
        weight_for_word = words
    else:
        # assume we're given an iterable of words and make a {word: count,...) dict
        weight_for_word = Counter(words)

    wc = WordCloud(**word_cloud_kwargs)
    wc.fit_words(weight_for_word)
    if save_filepath is not None:
        wc.to_file(save_filepath)
    return wc



"""Utils"""
import os
import re
from functools import partial, lru_cache

from urllib.parse import quote as url_quote
import requests


def get_pyversion():
    import sys

    _maj, _minor, *_ = sys.version_info
    return _maj, _minor


py_version = get_pyversion()

if py_version >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

proj_name, *_ = __name__.split('.')

proj_files = files(proj_name)
data_files = proj_files / 'data'

STOPWORDS = frozenset(
    map(str.strip, (data_files / 'stopwords.txt').read_text().split('\n'))
)


def conditional_print(*args, condition, **kwargs):
    """Print on `condition`

    Intended use:

    >>> verbose = True  # in the args of a function, and then in the function do:
    >>> _print = partial(conditional_print, condition=verbose)
    >>> # ... then use _print within function
    """
    if condition:
        print(*args, **kwargs)


def url_to_html(url, encoding='latin-1', **kwargs):
    try:
        r = requests.get(url, **kwargs)
        if r.status_code == 200:
            return r.content.decode(encoding=encoding)
    except Exception as e:
        print(f'ERROR with {url}: {e}')



def without_nones(d):
    return {k: v for k, v in d.items() if v is not None}


def print_status_code_and_content(response):
    print(f'ERROR: {response.status_code}: {response.content}')


def google_search_html(
    q, num=10, start=0, lr=None, on_error=print_status_code_and_content, **kwargs
):
    """
    Get's the html of google results (`start` through `start+num`) for query `q`.

    :param q: query
    :param num: the number of results you want (default is 10 -- docs say
        it's also the max, but haven't seen that to be true)
    :param start: Which result to start with (play with start and num to page through
        results
    :param lr: Language restriction (e.g. "lang_en" for English)
    :param on_error: if given, func that is called on response if status code not 200
    :param kwargs: Extra params:

    See https://developers.google.com/custom-search/v1/reference/rest/v1/cse/list
    for extra parameters and descriptions.


    """
    google_search_url = 'https://www.google.com/search'

    params = without_nones(dict(q=q, start=start, num=num, lr=lr, **kwargs))

    r = requests.get(google_search_url, params=params)
    if r.status_code:
        return r.content
    else:
        if on_error:
            return on_error(r)


def all_links_of_html(html, href_value_pattern='http.*'):
    from bs4 import BeautifulSoup
    href_value_pattern = re.compile(href_value_pattern)
    t = BeautifulSoup(html, features='lxml')
    return list(
        dict.fromkeys(  # this removes duplicates conserving order
            x.get('href')
            for x in t.find_all(attrs={'href': href_value_pattern}, recursive=True)
        )
    )


_google_url_prefix = '/url?q='
_google_url_junk = re.compile(r'&sa=\w+&ved=[\w-]+&usg=[\w-]+$')


def _google_url_post_proc(url):
    if url.startswith(_google_url_prefix):
        url = url[len(_google_url_prefix) :]
    if _google_url_junk.search(url):
        url = _google_url_junk.sub('', url)
    return url


def google_results_urls(html, postproc=_google_url_post_proc):
    return list(
        dict.fromkeys(  # to remove duplicates but keep order
            map(
                postproc,
                filter(lambda url: 'google.com' not in url, all_links_of_html(html)),
            )
        )
    )


DFLT_HTML_STORE_DIR = '~/odat/html'


def make_dflt_html_store_dir(store_dir=DFLT_HTML_STORE_DIR):
    store_dir = os.path.expanduser(store_dir)
    if not os.path.isdir(store_dir):
        answer = input(
            f'Can I make the directory to store htmls in them (Y/n): {store_dir}'
        )
        if answer.lower() in {'y', ''}:
            os.makedirs(store_dir)


def get_html_store(html_store=None):
    if html_store is None:
        html_store = os.path.expanduser(DFLT_HTML_STORE_DIR)
    if isinstance(html_store, str):
        make_dflt_html_store_dir(html_store)
        from py2store import LocalTextStore

        rootdir = html_store
        html_store = LocalTextStore(html_store, rootdir)
    return html_store

